"""Turms related functions

Turms is a python library for generating Turms operations from python classes.
This module contains functions that allow you to execute these operations using
rath.

As well as a plugin that plays well with turms and autoconnects the generated
code to rath, allowing you to execute typesfave operations using rath.

"""
